/*
 * Escape the Men's Room
 * Copyright (C) 2011 meatfighter.com
 *
 * This file is part of Escape the Men's Room.
 *
 * Escape the Men's Room is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Escape the Men's Room is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package escape;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

public class Main extends Applet implements Runnable {

  public static final int CANVAS_WIDTH = 800;
  public static final int CANVAS_HEIGHT = 600;
  public static final int MAX_FRAME_SKIPS = 8;

  public static Main main;

  public TitleMode titleMode;
  public UrinalMode urinalMode;
  public VentMode ventMode;
  public TrashCanMode trashCanMode;
  public TileMode tileMode;
  public PipesMode pipesMode;
  public DoorMode doorMode;
  public KeyholeMode keyholeMode;
  public KeycardMode keycardMode;
  public PaperTowelMode paperTowelMode;
  public LockMode lockMode;
  public SmokeAlarmMode smokeAlarmMode;
  public ConnectMode connectMode;
  public DrainMode drainMode;
  public DrainMagnetMode drainMagnetMode;
  public ClosetMode closetMode;
  public PrintoutMode printoutMode;
  public SinksMode sinksMode;
  public MirrorMode mirrorMode;
  public MirrorHoleMode mirrorHoleMode;
  public StallsMode stallsMode;
  public PennyUnlockMode pennyUnlockMode;
  public StallMode stallMode;
  public ToiletPaperMode toiletPaperMode;
  public KeypadMode keypadMode;
  public EndingMode endingMode;
  public CreditsMode creditsMode;

  public IMode mode;
  public IMode targetMode;
  public Stack<IMode> modeStack = new Stack<>();

  private Thread thread;
  private volatile boolean sleeping;
  private volatile boolean alive = true;

  public Main() {
    main = this;
    initializeModes();
  }

  private void initializeModes() {
    titleMode = new TitleMode();
    urinalMode = new UrinalMode();
    ventMode = new VentMode();
    trashCanMode = new TrashCanMode();
    tileMode = new TileMode();
    pipesMode = new PipesMode();
    doorMode = new DoorMode();
    keyholeMode = new KeyholeMode();
    keycardMode = new KeycardMode();
    paperTowelMode = new PaperTowelMode();
    lockMode = new LockMode();
    smokeAlarmMode = new SmokeAlarmMode();
    connectMode = new ConnectMode();
    drainMode = new DrainMode();
    drainMagnetMode = new DrainMagnetMode();
    closetMode = new ClosetMode();
    printoutMode = new PrintoutMode();
    sinksMode = new SinksMode();
    mirrorMode = new MirrorMode();
    mirrorHoleMode = new MirrorHoleMode();
    stallsMode = new StallsMode();
    pennyUnlockMode = new PennyUnlockMode();
    stallMode = new StallMode();
    toiletPaperMode = new ToiletPaperMode();
    keypadMode = new KeypadMode();
    endingMode = new EndingMode();
    creditsMode = new CreditsMode();

    mode = titleMode;
  }

  @Override
  public synchronized void start() {
    sleeping = false;
    notifyAll();
    if (thread == null) {      
      thread = new Thread(this);
      thread.start();
    }
  }

  @Override
  public synchronized void stop() {
    sleeping = true;
  }

  @Override
  public synchronized void destroy() {
    alive = false;
    sleeping = false;
    notifyAll();
  }

  @Override
  public void run() {

    Images.inventoryTile.getWidth(); // load images
    addMouseListener();

    BufferedImage image = new BufferedImage(CANVAS_WIDTH, CANVAS_HEIGHT,
        BufferedImage.TYPE_INT_RGB);
    Graphics2D g = (Graphics2D)image.getGraphics();
    Graphics2D g2 = null;
    int frameSkips = 0;

    long nextFrameStartTime = System.nanoTime();
    while(alive) {

      do {
        nextFrameStartTime += 16666667;
        frameSkips++;

        mode.update();

      } while(nextFrameStartTime < System.nanoTime()
          && frameSkips < MAX_FRAME_SKIPS);

      if (frameSkips >= MAX_FRAME_SKIPS) {
        nextFrameStartTime = System.nanoTime();
      }
      frameSkips = 0;

      if (targetMode != null) {
        mode = targetMode;
        mode.enter();
        targetMode = null;
        nextFrameStartTime = System.nanoTime();
        continue;
      }

      mode.render(g);

      // show the hidden buffer
      if (g2 == null) {
        g2 = (Graphics2D)getGraphics();
        requestFocus();
      } else {
        g2.drawImage(image, 0, 0, null);
      }

      // burn off extra cycles
      while(nextFrameStartTime - System.nanoTime() > 0) {
        Thread.yield();
      }

      // do not eat up CPU if user leaves browser
      if (sleeping) {
        synchronized(this) {
          while (sleeping && alive) {
            try {
              wait();
            } catch(Throwable t) {
            }
          }
        }
        nextFrameStartTime = System.nanoTime();
      }
    }
  }

  private void addMouseListener() {
    addMouseMotionListener(new MouseMotionListener() {
      @Override
      public void mouseDragged(MouseEvent e) {
        Mouse.x = e.getX();
        Mouse.y = e.getY();
      }
      @Override
      public void mouseMoved(MouseEvent e) {
        Mouse.x = e.getX();
        Mouse.y = e.getY();
      }
    });
    addMouseListener(new MouseAdapter() {
      @Override
      public void mousePressed(MouseEvent e) {
        Mouse.x = e.getX();
        Mouse.y = e.getY();
        Mouse.pressed = true;
      }
      @Override
      public void mouseReleased(MouseEvent e) {
        Mouse.x = e.getX();
        Mouse.y = e.getY();
        Mouse.pressed = false;
      }
      @Override
      public void mouseMoved(MouseEvent e) {
        Mouse.x = e.getX();
        Mouse.y = e.getY();
      }
      @Override
      public void mouseDragged(MouseEvent e) {
        Mouse.x = e.getX();
        Mouse.y = e.getY();
      }
    });
  }

  public void hideMouseCursor() {
    if (EventQueue.isDispatchThread()) {
      setCursor(Images.blankCursor);
    } else {
      Mouse.visible = false;
      EventQueue.invokeLater(new Runnable() {
        @Override
        public void run() {
          hideMouseCursor();
        }
      });
    }
  }

  public void showMouseCursor() {
    if (EventQueue.isDispatchThread()) {
      setCursor(Cursor.getDefaultCursor());
    } else {
      Mouse.visible = true;
      EventQueue.invokeLater(new Runnable() {
        @Override
        public void run() {
          showMouseCursor();
        }
      });
    }
  }

  public void pushMode(IMode mode) {
    modeStack.push(this.mode);
    setMode(mode);
  }

  public void popMode() {
    setMode(modeStack.pop());
  }

  public void setMode(IMode mode) {
    targetMode = mode;
  }

  // to run in window, uncomment below
  public static void main(String[] args) throws Throwable {
    javax.swing.JFrame frame = new javax.swing.JFrame("Escape the Men's Room");
    frame.setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
    Main applet = new Main();
    applet.setPreferredSize(
        new java.awt.Dimension(CANVAS_WIDTH, CANVAS_HEIGHT));
    frame.add(applet, java.awt.BorderLayout.CENTER);
    frame.setResizable(false);
    frame.pack();
    frame.setLocationRelativeTo(null);
    frame.setVisible(true);
    Thread.sleep(250);
    applet.start();
  }
}
